@@ -268,9 +268,11 @@ class _Transport(object):
                  retry_delay_growth=1.5,
                  retry_delay_jitter=0.1,
                  proxy=None,
-                 options=dict()):
+                 options=None):
         """
         """
+        if options is None:
+            options = dict()
         self.idx = idx
 
         self.type = kind
@@ -358,7 +360,7 @@ class Component(ObservableMixin):
     The factory of the session we will instantiate.
     """
 
-    def subscribe(self, topic, options=None):
+    def subscribe(self, topic, options=None, check_types=False):
         """
         A decorator as a shortcut for subscribing during on-join
 
@@ -376,12 +378,12 @@ class Component(ObservableMixin):
         def decorator(fn):
 
             def do_subscription(session, details):
-                return session.subscribe(fn, topic=topic, options=options)
+                return session.subscribe(fn, topic=topic, options=options, check_types=check_types)
             self.on('join', do_subscription)
             return fn
         return decorator
 
-    def register(self, uri, options=None):
+    def register(self, uri, options=None, check_types=False):
         """
         A decorator as a shortcut for registering during on-join
 
@@ -399,7 +401,7 @@ class Component(ObservableMixin):
         def decorator(fn):
 
             def do_registration(session, details):
-                return session.register(fn, procedure=uri, options=options)
+                return session.register(fn, procedure=uri, options=options, check_types=check_types)
             self.on('join', do_registration)
             return fn
         return decorator
@@ -415,7 +417,7 @@ class Component(ObservableMixin):
             pass a main() procedure, the session will not be closed
             (unless you arrange for .leave() to be called).
 
-        :type main: callable taking 2 args: reactor, ISession
+        :type main: callable taking two args ``reactor`` and ``ISession``
 
         :param transports: Transport configurations for creating
             transports. Each transport can be a WAMP URL, or a dict
@@ -443,25 +445,23 @@ class Component(ObservableMixin):
                 - ``options``: tbd
                 - ``proxy``: tbd
 
-        :type transports: None or unicode or list of dicts
+        :type transports: None or str or list
 
         :param realm: the realm to join
-        :type realm: unicode
+        :type realm: str
 
         :param authentication: configuration of authenticators
-        :type authentication: dict mapping auth_type to dict
+        :type authentication: dict
 
         :param session_factory: if None, ``ApplicationSession`` is
             used, otherwise a callable taking a single ``config`` argument
             that is used to create a new `ApplicationSession` instance.
-        :type session_factory: callable
 
         :param is_fatal: a callable taking a single argument, an
             ``Exception`` instance. The callable should return ``True`` if
             this error is "fatal", meaning we should not try connecting to
             the current transport again. The default behavior (on None) is
             to always return ``False``
-        :type is_fatal: callable taking one arg, or None
         """
         self.set_valid_events(
             [
@@ -723,7 +723,7 @@ class Component(ObservableMixin):
 
     def _connect_once(self, reactor, transport):
 
-        self.log.info(
+        self.log.debug(
             'connecting once using transport type "{transport_type}" '
             'over endpoint "{endpoint_desc}"',
             transport_type=transport.type,
@@ -767,7 +767,7 @@ class Component(ObservableMixin):
                         details=details,
                     )
                     if not txaio.is_called(done):
-                        if details.reason in ["wamp.close.normal"]:
+                        if details.reason in ["wamp.close.normal", "wamp.close.goodbye_and_out"]:
                             txaio.resolve(done, None)
                         else:
                             f = txaio.create_failure(
