@@ -37,14 +37,13 @@ from txaio import time_ns
 
 import cbor2
 import eth_keys
-# from eth_account import Account
 import nacl.secret
 import nacl.utils
 import nacl.public
 import txaio
-# import web3
 
-from ._util import hl, recover_eip712_signer, sign_eip712_data
+from ._util import hl, hlval
+from ._eip712_channel_close import sign_eip712_channel_close, recover_eip712_channel_close
 
 
 class KeySeries(object):
@@ -53,7 +52,7 @@ class KeySeries(object):
     and key offering (to the XBR market maker).
     """
 
-    def __init__(self, api_id, price, interval, on_rotate=None):
+    def __init__(self, api_id, price, interval=None, count=None, on_rotate=None):
         """
 
         :param api_id: ID of the API for which to generate keys.
@@ -62,20 +61,27 @@ class KeySeries(object):
         :param price: Price per key in key series.
         :type price: int
 
-        :param interval: Key rotation interval in seconds.
+        :param interval: Interval in seconds after which to auto-rotate key.
         :type interval: int
 
+        :param count: Number of encryption operations after which to auto-rotate key.
+        :type count: int
+
         :param on_rotate: Optional user callback fired after key was rotated.
         :type on_rotate: callable
         """
         assert type(api_id) == bytes and len(api_id) == 16
-        assert type(price) == int and price > 0
-        assert type(interval) == int and interval > 0
+        assert type(price) == int and price >= 0
+        assert interval is None or (type(interval) == int and interval > 0)
+        assert count is None or (type(count) == int and count > 0)
+        assert (interval is None and count is not None) or (interval is not None and count is None)
         assert on_rotate is None or callable(on_rotate)
 
         self._api_id = api_id
         self._price = price
         self._interval = interval
+        self._count = count
+        self._count_current = 0
         self._on_rotate = on_rotate
 
         self._id = None
@@ -94,7 +100,7 @@ class KeySeries(object):
         """
         return self._id
 
-    def encrypt(self, payload):
+    async def encrypt(self, payload):
         """
         Encrypt data with the current XBR data encryption key.
 
@@ -105,6 +111,13 @@ class KeySeries(object):
         :rtype: bytes
         """
         data = cbor2.dumps(payload)
+
+        if self._count is not None:
+            self._count_current += 1
+            if self._count_current >= self._count:
+                await self._rotate()
+                self._count_current = 0
+
         ciphertext = self._box.encrypt(data)
 
         return self._id, 'cbor', ciphertext
@@ -154,7 +167,7 @@ class KeySeries(object):
         # add key to archive
         self._archive[self._id] = (self._key, self._box)
 
-        self.log.info(
+        self.log.debug(
             '{tx_type} key "{key_id}" rotated [api_id="{api_id}"]',
             tx_type=hl('XBR ROTATE', color='magenta'),
             key_id=hl(uuid.UUID(bytes=self._id)),
@@ -167,8 +180,8 @@ class KeySeries(object):
 
 class PayingChannel(object):
     def __init__(self, adr, seq, balance):
-        assert type(adr) == bytes and len(adr) == 20
-        assert type(seq) == int and seq > 0
+        assert type(adr) == bytes and len(adr) == 16
+        assert type(seq) == int and seq >= 0
         assert type(balance) == int and balance >= 0
         self._adr = adr
         self._seq = seq
@@ -210,6 +223,7 @@ class SimpleSeller(object):
 
         # market maker address
         self._market_maker_adr = market_maker_adr
+        self._xbrmm_config = None
 
         # seller raw ethereum private key (32 bytes)
         self._pkey_raw = seller_key
@@ -261,7 +275,7 @@ class SimpleSeller(object):
         """
         return self._pkey.public_key
 
-    def add(self, api_id, prefix, price, interval, categories=None):
+    def add(self, api_id, prefix, price, interval=None, count=None, categories=None):
         """
         Add a new (rotating) private encryption key for encrypting data on the given API.
 
@@ -271,12 +285,17 @@ class SimpleSeller(object):
         :param price: Price in XBR token per key.
         :type price: int
 
-        :param interval: Interval (in seconds) in which to auto-rotate the encryption key.
+        :param interval: Interval (in seconds) after which to auto-rotate the encryption key.
         :type interval: int
+
+        :param count: Number of encryption operations after which to auto-rotate the encryption key.
+        :type count: int
         """
         assert type(api_id) == bytes and len(api_id) == 16 and api_id not in self._keys
-        assert type(price) == int and price > 0
-        assert type(interval) == int and interval > 0
+        assert type(price) == int and price >= 0
+        assert interval is None or (type(interval) == int and interval > 0)
+        assert count is None or (type(count) == int and count > 0)
+        assert (interval is None and count is not None) or (interval is not None and count is None)
         assert categories is None or (type(categories) == dict and (type(k) == str for k in categories.keys()) and (type(v) == str for v in categories.values())), 'invalid categories type (must be dict) or category key or value type (must both be string)'
 
         async def on_rotate(key_series):
@@ -311,7 +330,7 @@ class SimpleSeller(object):
                                                      copies=None,
                                                      provider_id=provider_id)
 
-                    self.log.info(
+                    self.log.debug(
                         '{tx_type} key "{key_id}" offered for {price} [api_id={api_id}, prefix="{prefix}", delegate="{delegate}"]',
                         tx_type=hl('XBR OFFER ', color='magenta'),
                         key_id=hl(uuid.UUID(bytes=key_id)),
@@ -340,9 +359,9 @@ class SimpleSeller(object):
                 self.log.warn('Failed to place offer for key! Retrying {retries}/5 ..', retries=retries)
                 await asyncio.sleep(1)
 
-        key_series = self.KeySeries(api_id, price, interval, on_rotate)
+        key_series = self.KeySeries(api_id, price, interval=interval, count=count, on_rotate=on_rotate)
         self._keys[api_id] = key_series
-        self.log.info('Created new key series {key_series}', key_series=key_series)
+        self.log.debug('Created new key series {key_series}', key_series=key_series)
 
         return key_series
 
@@ -360,9 +379,18 @@ class SimpleSeller(object):
         self._session = session
         self._session_regs = []
 
-        self.log.info('Start selling from seller delegate address {address} (public key 0x{public_key}..)',
-                      address=hl(self._caddr),
-                      public_key=binascii.b2a_hex(self._pkey.public_key[:10]).decode())
+        self.log.debug('Start selling from seller delegate address {address} (public key 0x{public_key}..)',
+                       address=hl(self._caddr),
+                       public_key=binascii.b2a_hex(self._pkey.public_key[:10]).decode())
+
+        # get the currently active (if any) paying channel for the delegate
+        self._channel = await session.call('xbr.marketmaker.get_active_paying_channel', self._addr)
+        if not self._channel:
+            raise Exception('no active paying channel found')
+
+        channel_oid = self._channel['channel_oid']
+        assert type(channel_oid) == bytes and len(channel_oid) == 16
+        self._channel_oid = uuid.UUID(bytes=channel_oid)
 
         procedure = 'xbr.provider.{}.sell'.format(self._provider_id)
         reg = await session.register(self.sell, procedure, options=RegisterOptions(details_arg='details'))
@@ -377,28 +405,29 @@ class SimpleSeller(object):
         for key_series in self._keys.values():
             await key_series.start()
 
-        # get the currently active (if any) paying channel for the delegate
-        channel = await session.call('xbr.marketmaker.get_active_paying_channel', self._addr)
+        self._xbrmm_config = await session.call('xbr.marketmaker.get_config')
 
         # get the current (off-chain) balance of the paying channel
-        paying_balance = await session.call('xbr.marketmaker.get_paying_channel_balance', channel['channel'])
+        paying_balance = await session.call('xbr.marketmaker.get_paying_channel_balance', self._channel_oid.bytes)
         # FIXME
         if type(paying_balance['remaining']) == bytes:
             paying_balance['remaining'] = unpack_uint256(paying_balance['remaining'])
 
-        self.log.info('Delegate has currently active paying channel address {paying_channel_adr}',
-                      paying_channel_adr=hl('0x' + binascii.b2a_hex(channel['channel']).decode()))
+        if not paying_balance['remaining'] > 0:
+            raise Exception('no off-chain balance remaining on paying channel')
 
-        self._channels[channel['channel']] = PayingChannel(channel['channel'], paying_balance['seq'], paying_balance['remaining'])
+        self._channels[channel_oid] = PayingChannel(channel_oid, paying_balance['seq'], paying_balance['remaining'])
         self._state = SimpleSeller.STATE_STARTED
 
         # FIXME
-        self._channel = channel
         self._balance = paying_balance['remaining']
         if type(self._balance) == bytes:
             self._balance = unpack_uint256(self._balance)
         self._seq = paying_balance['seq']
 
+        self.log.info('Ok, seller delegate started [active paying channel {channel_oid} with remaining balance {remaining} at sequence {seq}]',
+                      channel_oid=hl(self._channel_oid), remaining=hlval(self._balance), seq=hlval(self._seq))
+
         return paying_balance['remaining']
 
     async def stop(self):
@@ -432,6 +461,8 @@ class SimpleSeller(object):
             self._state = SimpleSeller.STATE_STOPPED
             self._session = None
 
+        self.log.info('Ok, seller delegate stopped.')
+
     async def balance(self):
         """
         Return current (off-chain) balance of paying channel:
@@ -448,7 +479,7 @@ class SimpleSeller(object):
         if not self._session or not self._session.is_attached():
             raise RuntimeError('market-maker session not attached')
 
-        paying_balance = await self._session.call('xbr.marketmaker.get_payment_channel_balance', self._channel['channel'])
+        paying_balance = await self._session.call('xbr.marketmaker.get_paying_channel_balance', self._channel['channel_oid'])
 
         return paying_balance
 
@@ -474,17 +505,17 @@ class SimpleSeller(object):
 
         keyseries = self._keys[api_id]
 
-        key_id, serializer, ciphertext = keyseries.encrypt(payload)
+        key_id, serializer, ciphertext = await keyseries.encrypt(payload)
 
         return key_id, serializer, ciphertext
 
-    def close_channel(self, market_maker_adr, channel_adr, channel_seq, channel_balance, channel_is_final,
+    def close_channel(self, market_maker_adr, channel_oid, channel_seq, channel_balance, channel_is_final,
                       marketmaker_signature, details=None):
         """
         Called by a XBR Market Maker to close a paying channel.
         """
         assert type(market_maker_adr) == bytes and len(market_maker_adr) == 20, 'market_maker_adr must be bytes[20], but was {}'.format(type(market_maker_adr))
-        assert type(channel_adr) == bytes and len(channel_adr) == 20, 'channel_adr must be bytes[20], but was {}'.format(type(channel_adr))
+        assert type(channel_oid) == bytes and len(channel_oid) == 16, 'channel_oid must be bytes[16], but was {}'.format(type(channel_oid))
         assert type(channel_seq) == int, 'channel_seq must be int, but was {}'.format(type(channel_seq))
         assert type(channel_balance) == bytes and len(channel_balance) == 32, 'channel_balance must be bytes[32], but was {}'.format(type(channel_balance))
         assert type(channel_is_final) == bool, 'channel_is_final must be bool, but was {}'.format(type(channel_is_final))
@@ -497,10 +528,10 @@ class SimpleSeller(object):
                                    '{}.sell() - unexpected market maker (delegate) address: expected 0x{}, but got 0x{}'.format(self.__class__.__name__, binascii.b2a_hex(self._market_maker_adr).decode(), binascii.b2a_hex(market_maker_adr).decode()))
 
         # FIXME: must be the currently active channel .. and we need to track all of these
-        if channel_adr != self._channel['channel']:
+        if channel_oid != self._channel['channel_oid']:
             self._session.leave()
-            raise ApplicationError('xbr.error.unexpected_channel_adr',
-                                   '{}.sell() - unexpected paying channel address: expected 0x{}, but got 0x{}'.format(self.__class__.__name__, binascii.b2a_hex(self._channel['channel']).decode(), binascii.b2a_hex(channel_adr).decode()))
+            raise ApplicationError('xbr.error.unexpected_channel_oid',
+                                   '{}.sell() - unexpected paying channel address: expected 0x{}, but got 0x{}'.format(self.__class__.__name__, binascii.b2a_hex(self._channel['channel_oid']).decode(), binascii.b2a_hex(channel_oid).decode()))
 
         # channel sequence number: check we have consensus on off-chain channel state with peer (which is the market maker)
         if channel_seq != self._seq:
@@ -514,7 +545,7 @@ class SimpleSeller(object):
                                    '{}.sell() - unexpected channel (after tx) balance: expected {}, but got {}'.format(self.__class__.__name__, self._balance, channel_balance))
 
         # XBRSIG: check the signature (over all input data for the buying of the key)
-        signer_address = recover_eip712_signer(channel_adr, channel_seq, channel_balance, channel_is_final, marketmaker_signature)
+        signer_address = recover_eip712_channel_close(channel_oid, channel_seq, channel_balance, channel_is_final, marketmaker_signature)
         if signer_address != market_maker_adr:
             self.log.warn('{klass}.sell()::XBRSIG[4/8] - EIP712 signature invalid: signer_address={signer_address}, delegate_adr={delegate_adr}',
                           klass=self.__class__.__name__,
@@ -523,7 +554,7 @@ class SimpleSeller(object):
             raise ApplicationError('xbr.error.invalid_signature', '{}.sell()::XBRSIG[4/8] - EIP712 signature invalid or not signed by market maker'.format(self.__class__.__name__))
 
         # XBRSIG: compute EIP712 typed data signature
-        seller_signature = sign_eip712_data(self._pkey_raw, channel_adr, channel_seq, channel_balance, channel_is_final)
+        seller_signature = sign_eip712_channel_close(self._pkey_raw, channel_oid, channel_seq, channel_balance, channel_is_final)
 
         receipt = {
             'delegate': self._addr,
@@ -533,18 +564,18 @@ class SimpleSeller(object):
             'signature': seller_signature,
         }
 
-        self.log.info('{klass}.close_channel() - {tx_type} closing channel {channel_adr}, closing balance {channel_balance}, closing sequence {channel_seq} [caller={caller}, caller_authid="{caller_authid}"]',
-                      klass=self.__class__.__name__,
-                      tx_type=hl('XBR CLOSE  ', color='magenta'),
-                      channel_balance=hl(str(int(channel_balance / 10 ** 18)) + ' XBR', color='magenta'),
-                      channel_seq=hl(channel_seq),
-                      channel_adr=hl(binascii.b2a_hex(channel_adr).decode()),
-                      caller=hl(details.caller),
-                      caller_authid=hl(details.caller_authid))
+        self.log.debug('{klass}.close_channel() - {tx_type} closing channel {channel_oid}, closing balance {channel_balance}, closing sequence {channel_seq} [caller={caller}, caller_authid="{caller_authid}"]',
+                       klass=self.__class__.__name__,
+                       tx_type=hl('XBR CLOSE  ', color='magenta'),
+                       channel_balance=hl(str(int(channel_balance / 10 ** 18)) + ' XBR', color='magenta'),
+                       channel_seq=hl(channel_seq),
+                       channel_oid=hl(binascii.b2a_hex(channel_oid).decode()),
+                       caller=hl(details.caller),
+                       caller_authid=hl(details.caller_authid))
 
         return receipt
 
-    def sell(self, market_maker_adr, buyer_pubkey, key_id, channel_adr, channel_seq, amount, balance, signature, details=None):
+    def sell(self, market_maker_adr, buyer_pubkey, key_id, channel_oid, channel_seq, amount, balance, signature, details=None):
         """
         Called by a XBR Market Maker to buy a data encyption key. The XBR Market Maker here is
         acting for (triggered by) the XBR buyer delegate.
@@ -559,8 +590,8 @@ class SimpleSeller(object):
         :param key_id: The UUID of the data encryption key to buy.
         :type key_id: bytes of length 16
 
-        :param channel_adr: The on-chain channel contract address.
-        :type channel_adr: bytes of length 20
+        :param channel_oid: The on-chain channel contract address.
+        :type channel_oid: bytes of length 16
 
         :param channel_seq: Paying channel sequence off-chain transaction number.
         :type channel_seq: int
@@ -585,7 +616,7 @@ class SimpleSeller(object):
         assert type(market_maker_adr) == bytes and len(market_maker_adr) == 20, 'delegate_adr must be bytes[20]'
         assert type(buyer_pubkey) == bytes and len(buyer_pubkey) == 32, 'buyer_pubkey must be bytes[32]'
         assert type(key_id) == bytes and len(key_id) == 16, 'key_id must be bytes[16]'
-        assert type(channel_adr) == bytes and len(channel_adr) == 20, 'channel_adr must be bytes[20]'
+        assert type(channel_oid) == bytes and len(channel_oid) == 16, 'channel_oid must be bytes[16]'
         assert type(channel_seq) == int, 'channel_seq must be int'
         assert type(amount) == bytes and len(amount) == 32, 'amount_paid must be bytes[32], but was {}'.format(type(amount))
         assert type(balance) == bytes and len(amount) == 32, 'post_balance must be bytes[32], but was {}'.format(type(balance))
@@ -606,10 +637,10 @@ class SimpleSeller(object):
         key_series = self._keys_map[key_id]
 
         # FIXME: must be the currently active channel .. and we need to track all of these
-        if channel_adr != self._channel['channel']:
+        if channel_oid != self._channel['channel_oid']:
             self._session.leave()
-            raise ApplicationError('xbr.error.unexpected_channel_adr',
-                                   '{}.sell() - unexpected paying channel address: expected 0x{}, but got 0x{}'.format(self.__class__.__name__, binascii.b2a_hex(self._channel['channel']).decode(), binascii.b2a_hex(channel_adr).decode()))
+            raise ApplicationError('xbr.error.unexpected_channel_oid',
+                                   '{}.sell() - unexpected paying channel address: expected 0x{}, but got 0x{}'.format(self.__class__.__name__, binascii.b2a_hex(self._channel['channel_oid']).decode(), binascii.b2a_hex(channel_oid).decode()))
 
         # channel sequence number: check we have consensus on off-chain channel state with peer (which is the market maker)
         if channel_seq != self._seq + 1:
@@ -621,8 +652,16 @@ class SimpleSeller(object):
             raise ApplicationError('xbr.error.unexpected_channel_balance',
                                    '{}.sell() - unexpected channel (after tx) balance: expected {}, but got {}'.format(self.__class__.__name__, self._balance - amount, balance))
 
+        # FIXME
+        current_block_number = 1
+        verifying_chain_id = self._xbrmm_config['verifying_chain_id']
+        verifying_contract_adr = binascii.a2b_hex(self._xbrmm_config['verifying_contract_adr'][2:])
+
+        market_oid = self._channel['market_oid']
+
         # XBRSIG[4/8]: check the signature (over all input data for the buying of the key)
-        signer_address = recover_eip712_signer(channel_adr, channel_seq, balance, False, signature)
+        signer_address = recover_eip712_channel_close(verifying_chain_id, verifying_contract_adr, current_block_number,
+                                                      market_oid, channel_oid, channel_seq, balance, False, signature)
         if signer_address != market_maker_adr:
             self.log.warn('{klass}.sell()::XBRSIG[4/8] - EIP712 signature invalid: signer_address={signer_address}, delegate_adr={delegate_adr}',
                           klass=self.__class__.__name__,
@@ -641,7 +680,9 @@ class SimpleSeller(object):
         assert type(sealed_key) == bytes and len(sealed_key) == 80, '{}.sell() - unexpected sealed key computed (expected bytes[80]): {}'.format(self.__class__.__name__, sealed_key)
 
         # XBRSIG[5/8]: compute EIP712 typed data signature
-        seller_signature = sign_eip712_data(self._pkey_raw, self._channel['channel'], self._seq, self._balance)
+        seller_signature = sign_eip712_channel_close(self._pkey_raw, verifying_chain_id, verifying_contract_adr,
+                                                     current_block_number, market_oid, channel_oid, self._seq,
+                                                     self._balance, False)
 
         receipt = {
             # key ID that has been bought
